@@ -1,5 +1,5 @@
 #
-# Copyright (c), 2022, SISSA (International School for Advanced Studies).
+# Copyright (c), 2022-2026, SISSA (International School for Advanced Studies).
 # All rights reserved.
 # This file is distributed under the terms of the MIT License.
 # See the file 'LICENSE' in the root directory of the present
@@ -68,7 +68,8 @@ def get_serialization_params(params: Union[None, ElementNode, XPathMap] = None,
                 # TODO: doesn't work within element nodes
                 if isinstance(value, XPathArray):
                     value = value.items()
-                if not isinstance(value, list) or not all(isinstance(x, QName) for x in value):
+                if not isinstance(value, list) or \
+                        not all(isinstance(x, QName) for x in value):
                     raise xpath_error('XPTY0004', token=token)
                 kwargs['cdata_section'] = value
 
@@ -139,7 +140,7 @@ def get_serialization_params(params: Union[None, ElementNode, XPathMap] = None,
                 kwargs[key] = value
 
     elif isinstance(params, ElementNode):
-        root = cast(Union[EtreeElementProtocol, LxmlElementProtocol], params.obj)
+        root = cast(Union[EtreeElementProtocol, LxmlElementProtocol], params.value)
         if root.tag != SERIALIZATION_PARAMS:
             msg = 'output:serialization-parameters tag expected'
             raise xpath_error('XPTY0004', msg, token)
@@ -293,14 +294,14 @@ def serialize_to_xml(elements: Iterable[Any],
     for item in iter_normalized(elements, item_separator):
         if isinstance(item, ElementNode):
             assert isinstance(item, EtreeElementNode)
-            elem = item.obj
+            elem = item.value
         elif isinstance(item, (AttributeNode, NamespaceNode)):
             raise xpath_error('SENR0001', token=token)
         elif isinstance(item, TextNode):
             if item.parent is not None and item.parent.name in cdata_section:
-                chunks.append(f'<![CDATA[{item.obj}]]>')
+                chunks.append(f'<![CDATA[{item.value}]]>')
             else:
-                chunks.append(item.obj)
+                chunks.append(item.value)
             continue
         elif not isinstance(item, str):
             raise xpath_error('SENR0001', token=token)
@@ -352,7 +353,7 @@ def serialize_to_json(elements: Iterable[Any],
                     return ''.join(self.default(child) for child in obj)
                 elif isinstance(obj, ElementNode):
                     assert isinstance(obj, EtreeElementNode)
-                    elem = obj.obj
+                    elem = obj.value
                     assert etree_module is not None
 
                     try:
@@ -368,15 +369,12 @@ def serialize_to_json(elements: Iterable[Any],
                 elif isinstance(obj, (AttributeNode, NamespaceNode)):
                     return f'{obj.name}="{obj.string_value}"'
                 elif isinstance(obj, TextNode):
-                    return obj.obj
+                    return obj.value
                 elif isinstance(obj, CommentNode):
                     return f'<!--{obj.string_value}-->'
                 else:
                     return f'<?{obj.name} {obj.string_value}?>'
             elif isinstance(obj, XPathMap):
-                if any(isinstance(v, list) and len(v) > 1 for v in obj.values()):
-                    raise xpath_error('SERE0023', token=token)
-
                 map_keys = set()
                 map_items = []
                 k: Any
@@ -397,6 +395,8 @@ def serialize_to_json(elements: Iterable[Any],
                 return str(obj)
             elif isinstance(obj, Decimal):
                 return float(Decimal(obj).quantize(Decimal("0.01"), ROUND_UP))
+            elif isinstance(obj, list):
+                return [v for v in obj]
             else:
                 return super().default(obj)
 
