@@ -1,5 +1,5 @@
 #
-# Copyright (c), 2018-2021, SISSA (International School for Advanced Studies).
+# Copyright (c), 2018-2026, SISSA (International School for Advanced Studies).
 # All rights reserved.
 # This file is distributed under the terms of the MIT License.
 # See the file 'LICENSE' in the root directory of the present
@@ -15,20 +15,21 @@ from abc import ABCMeta
 import locale
 from collections.abc import Callable, MutableMapping
 from urllib.parse import urlparse
-from typing import cast, Any, ClassVar, Optional, Union
+from typing import Any, ClassVar, cast
 
-from elementpath.aliases import NamespacesType, NargsType
-from elementpath.helpers import upper_camel_case, is_ncname, ordinal
-from elementpath.exceptions import ElementPathError, ElementPathTypeError, \
-    ElementPathValueError, MissingContextError, xpath_error
-from elementpath.namespaces import XSD_NAMESPACE, XML_NAMESPACE, \
-    XPATH_FUNCTIONS_NAMESPACE, XQT_ERRORS_NAMESPACE, \
-    XSD_NOTATION, XSD_ANY_ATOMIC_TYPE, get_prefixed_name
+import elementpath.aliases as ta
+
+from elementpath.helpers import upper_camel_case, is_ncname
+from elementpath.exceptions import ElementPathTypeError, ElementPathValueError, xpath_error
+from elementpath.namespaces import XML_NAMESPACE, XSD_NAMESPACE, XPATH_FUNCTIONS_NAMESPACE, \
+    XQT_ERRORS_NAMESPACE, XSD_NOTATION, XSD_ANY_ATOMIC_TYPE
+from elementpath.namespaces import get_prefixed_name
+from elementpath.datatypes import QName, builtin_atomic_types
 from elementpath.collations import UNICODE_COLLATION_BASE_URI, UNICODE_CODEPOINT_COLLATION
-from elementpath.datatypes import UntypedAtomic, AtomicType, QName
-from elementpath.xpath_tokens import XPathToken, ProxyToken, XPathFunction, XPathConstructor
-from elementpath.xpath_context import XPathContext, XPathSchemaContext
-from elementpath.sequence_types import is_sequence_type, match_sequence_type
+from elementpath.xpath_tokens import XPathToken, ProxyToken, XPathFunction, \
+    XPathConstructor, SchemaConstructor, ExternalFunction
+from elementpath.sequence_types import get_function_signatures, is_sequence_type, \
+    match_sequence_type
 from elementpath.schema_proxy import AbstractSchemaProxy
 from elementpath.xpath1 import XPath1Parser
 
@@ -98,18 +99,18 @@ class XPath2Parser(XPath1Parser):
     token: XPathToken
     next_token: XPathToken
 
-    def __init__(self, namespaces: Optional[NamespacesType] = None,
+    def __init__(self, namespaces: ta.NamespacesType | None = None,
                  strict: bool = True,
                  compatibility_mode: bool = False,
-                 default_collation: Optional[str] = None,
-                 default_namespace: Optional[str] = None,
-                 function_namespace: Optional[str] = None,
-                 xsd_version: Optional[str] = None,
-                 schema: Optional[AbstractSchemaProxy] = None,
-                 base_uri: Optional[str] = None,
-                 variable_types: Optional[dict[str, str]] = None,
-                 document_types: Optional[dict[str, str]] = None,
-                 collection_types: Optional[NamespacesType] = None,
+                 default_collation: str | None = None,
+                 default_namespace: str | None = None,
+                 function_namespace: str | None = None,
+                 xsd_version: str | None = None,
+                 schema: AbstractSchemaProxy | None = None,
+                 base_uri: str | None = None,
+                 variable_types: dict[str, str] | None = None,
+                 document_types: dict[str, str] | None = None,
+                 collection_types: ta.NamespacesType | None = None,
                  default_collection_type: str = 'node()*') -> None:
 
         super(XPath2Parser, self).__init__(namespaces, strict)
@@ -215,7 +216,7 @@ class XPath2Parser(XPath1Parser):
         except (AttributeError, NotImplementedError):
             return self._xsd_version
 
-    def advance(self, *symbols: str,  message: Optional[str] = None) -> XPathToken:
+    def advance(self, *symbols: str,  message: str | None = None) -> XPathToken:
         super(XPath2Parser, self).advance(*symbols, message=message)
 
         if self.next_token.symbol == '(:':
@@ -241,131 +242,99 @@ class XPath2Parser(XPath1Parser):
         return self.token
 
     @classmethod
-    def constructor(cls, symbol: str, bp: int = 90, nargs: NargsType = 1,
-                    sequence_types: Union[tuple[()], tuple[str, ...], list[str]] = (),
-                    label: Union[str, tuple[str, ...]] = 'constructor function') \
+    def constructor(cls, symbol: str, bp: int = 90, nargs: ta.NargsType = 1,
+                    sequence_types: tuple[()] | tuple[str, ...] | list[str] = (),
+                    label: str | tuple[str, ...] = 'constructor function') \
             -> Callable[[Callable[..., Any]], Callable[..., Any]]:
         """
         Statically creates a constructor token class, that is registered in the globals
         of the module where the method is called.
         """
-        def nud_(self: XPathConstructor) -> XPathConstructor:
-            if not self.parser.parse_arguments:
-                return self
-
-            try:
-                self.parser.advance('(')
-                self[0:] = self.parser.expression(5),
-                if self.parser.next_token.symbol == ',':
-                    msg = 'Too many arguments: expected at most 1 argument'
-                    raise self.error('XPST0017', msg)
-                self.parser.advance(')')
-            except SyntaxError:
-                raise self.error('XPST0017') from None
-            else:
-                if self[0].symbol == '?':
-                    self.to_partial_function()
-                return self
-
-        def evaluate_(self: XPathConstructor, context: Optional[XPathContext] = None) \
-                -> Union[list[None], AtomicType]:
-            if self.context is not None:
-                context = self.context
-
-            arg = self.data_value(self.get_argument(context))
-            if arg is None:
-                return []
-            elif arg == '?' and self[0].symbol == '?':
-                raise self.error('XPTY0004', "cannot evaluate a partial function")
-
-            try:
-                if isinstance(arg, UntypedAtomic):
-                    return self.cast(arg.value)
-                return self.cast(arg)
-            except ElementPathError:
-                raise
-            except (TypeError, ValueError) as err:
-                if isinstance(context, XPathSchemaContext):
-                    return []
-                raise self.error('FORG0001', err) from None
-
         if not sequence_types:
             assert nargs == 1
             sequence_types = ('xs:anyAtomicType?', 'xs:%s?' % symbol)
 
-        token_class = cls.register(symbol, nargs=nargs, sequence_types=sequence_types,
-                                   label=label, bases=(XPathConstructor,), lbp=bp, rbp=bp,
-                                   nud=nud_, evaluate=evaluate_)
+        type_name = f'{{{XSD_NAMESPACE}}}{symbol}'
+        kwargs = {
+            'bases': (XPathConstructor,),
+            'label': label,
+            'nargs': nargs,
+            'lbp': bp,
+            'rbp': bp,
+            'sequence_types': sequence_types,
+            'name': type_name,
+            'type_class': builtin_atomic_types.get(type_name),
+        }
+        token_class = cls.register(symbol, **kwargs)
 
         def bind(func: Callable[..., Any]) -> Callable[..., Any]:
             method_name = func.__name__.partition('_')[0]
             if method_name != 'cast':
-                raise ValueError("The function name must be 'cast' or starts with 'cast_'")
+                raise ValueError("The function name must be 'cast' or starting with 'cast_'")
             setattr(token_class, method_name, func)
             return func
         return bind
 
-    def schema_constructor(self, atomic_type_name: str, bp: int = 90) \
-            -> type[XPathFunction]:
-        """Dynamically registers a token class for a schema atomic type constructor function."""
+    def _register(self, symbol: str, class_name: str, **kwargs: Any) \
+            -> type[ta.XPathTokenType]:
+        """
+        Register/update a token class in the symbol table of a parser instance.
+
+        :param symbol: The identifier symbol for the new class.
+        :param class_name: The name of the new class.
+        :param kwargs: Optional attributes/methods for the token class.
+        :return: A token class.
+        """
+        kwargs['symbol'] = symbol
+        if 'lookup_name' not in kwargs:
+            lookup_name = kwargs['lookup_name'] = symbol
+        else:
+            lookup_name = kwargs['lookup_name']
+
+        if 'label' not in kwargs:
+            kwargs['label'] = 'symbol'
+
+        token_class_bases = kwargs.get('bases', (self.token_base_class,))
+        kwargs.update({
+            '__module__': self.__module__,
+            '__qualname__': class_name,
+            '__return__': None
+        })
+        token_class = cast(
+            type[ta.XPathTokenType],
+            cast(object, ABCMeta(class_name, token_class_bases, kwargs))
+        )
+
+        if self.symbol_table is self.__class__.symbol_table:
+            self.symbol_table = copy.copy(self.symbol_table)
+        self.symbol_table[lookup_name] = token_class
+        self.tokenizer = None
+        return token_class
+
+    def schema_constructor(self, atomic_type_name: str, bp: int = 90) -> type[XPathFunction]:
+        """
+        Dynamically register a token class for a schema atomic type constructor function.
+        """
         if atomic_type_name in (XSD_ANY_ATOMIC_TYPE, XSD_NOTATION):
             raise xpath_error('XPST0080')
 
-        def nud_(self_: XPathFunction) -> XPathFunction:
-            self_.parser.advance('(')
-            self_[0:] = self_.parser.expression(5),
-            self_.parser.advance(')')
-
-            try:
-                self_.evaluate()  # for static context evaluation
-            except MissingContextError:
-                pass
-            return self_
-
-        def evaluate_(self_: XPathFunction, context: Optional[XPathContext] = None) \
-                -> Union[list[None], AtomicType]:
-            arg = self_.get_argument(context)
-            if arg is None or self_.parser.schema is None:
-                return []
-
-            value = self_.string_value(arg)
-            try:
-                return self_.parser.schema.cast_as(value, atomic_type_name)
-            except (TypeError, ValueError) as err:
-                if isinstance(context, XPathSchemaContext):
-                    return []
-                raise self_.error('FORG0001', err)
-
         symbol = get_prefixed_name(atomic_type_name, self.namespaces)
-        token_class_name = "_%sConstructorFunction" % symbol.replace(':', '_')
+        class_name = "_%sConstructorFunction" % symbol.replace(':', '_')
         kwargs = {
-            'symbol': symbol,
+            'bases': (SchemaConstructor,),
+            'name': atomic_type_name,
             'nargs': 1,
             'label': 'constructor function',
             'pattern': r'\b%s(?=\s*\(|\s*\(\:.*\:\)\()' % symbol,
             'lbp': bp,
             'rbp': bp,
-            'nud': nud_,
-            'evaluate': evaluate_,
-            '__module__': self.__module__,
-            '__qualname__': token_class_name,
-            '__return__': None
         }
-        token_class = cast(
-            type[XPathFunction], ABCMeta(token_class_name, (XPathFunction,), kwargs)
-        )
-
-        if self.symbol_table is self.__class__.symbol_table:
-            self.symbol_table = copy.copy(self.symbol_table)
-        self.symbol_table[symbol] = token_class
-        self.tokenizer = None
-
-        return token_class
+        return cast(type[XPathFunction], self._register(symbol, class_name, **kwargs))
 
     def external_function(self,
                           callback: Callable[..., Any],
-                          name: Optional[str] = None,
-                          prefix: Optional[str] = None,
+                          name: str | None = None,
+                          prefix: str | None = None,
                           sequence_types: tuple[str, ...] = (),
                           bp: int = 90) -> type[XPathFunction]:
         """Registers a token class for an external function."""
@@ -377,7 +346,7 @@ class XPath2Parser(XPath1Parser):
         elif symbol in self.RESERVED_FUNCTION_NAMES:
             raise ElementPathValueError(f'{symbol!r} is a reserved function name')
 
-        nargs: NargsType
+        nargs: ta.NargsType
         spec = inspect.getfullargspec(callback)
         if spec.varargs is not None:
             if spec.args:
@@ -396,117 +365,59 @@ class XPath2Parser(XPath1Parser):
             namespace = XPATH_FUNCTIONS_NAMESPACE
             qname = QName(XPATH_FUNCTIONS_NAMESPACE, f'fn:{symbol}')
 
-        class_name = f'{upper_camel_case(qname.qname)}ExternalFunction'
-        lookup_name = qname.expanded_name
+        function_signatures = get_function_signatures(qname, nargs, sequence_types)
 
-        if self.symbol_table is self.__class__.symbol_table:
-            self.symbol_table = copy.copy(self.symbol_table)
-
-        if lookup_name in self.symbol_table:
+        if qname.expanded_name in self.symbol_table:
             msg = f'function {qname.qname!r} is already registered'
             raise ElementPathValueError(msg)
-        elif symbol not in self.symbol_table or \
-                not issubclass(self.symbol_table[symbol], ProxyToken):
-
-            if symbol in self.symbol_table:
-                token_cls = self.symbol_table[symbol]
-                if not issubclass(token_cls, XPathFunction) \
-                        or token_cls.label == 'kind test':
-                    msg = f'{symbol!r} name collides with {token_cls!r}'
-                    raise ElementPathValueError(msg)
-                if namespace == token_cls.namespace:
-                    msg = f'function {qname.qname!r} is already registered'
-                    raise ElementPathValueError(msg)
-
-                # Move the token class before register the proxy token
-                self.symbol_table[f'{{{token_cls.namespace}}}{symbol}'] = token_cls
-
-            token_class_name = f'{upper_camel_case(qname.local_name)}FunctionProxy'
-            kwargs = {
-                'class_name': token_class_name,
-                'symbol': symbol,
-                'label': 'function',
-                'lbp': bp,
-                'rbp': bp,
-                '__module__': self.__module__,
-                '__qualname__': token_class_name,
-                '__return__': None
-            }
-            self.symbol_table[symbol] = cast(
-                type[ProxyToken], ABCMeta(class_name, (ProxyToken,), kwargs)
-            )
-
-        def evaluate_external_function(self_: XPathFunction,
-                                       context: Optional[XPathContext] = None) -> Any:
-            args = []
-            for k in range(len(self_)):
-                try:
-                    if sequence_types[k][-1] in '+*':
-                        arg = self_[k].evaluate(context)
-                    else:
-                        arg = self_.get_argument(context, index=k)
-                except IndexError:
-                    arg = self_.get_argument(context, index=k)
-                args.append(arg)
-
-            if sequence_types:
-                for k, (arg, st) in enumerate(zip(args, sequence_types), start=1):
-                    if not match_sequence_type(arg, st, self):
-                        msg_ = f"{ordinal(k)} argument does not match sequence type {st!r}"
-                        raise xpath_error('XPDY0050', msg_)
-
-                result = callback(*args)
-                if not match_sequence_type(result, sequence_types[-1], self):
-                    msg_ = f"Result does not match sequence type {sequence_types[-1]!r}"
-                    raise xpath_error('XPDY0050', msg_)
-                return result
-
-            return callback(*args)
+        elif symbol not in self.symbol_table:
+            lookup_name = symbol
+        else:
+            lookup_name = qname.expanded_name
+            token_class = self.symbol_table[symbol]
+            if issubclass(token_class, ProxyToken):
+                pass
+            elif namespace == token_class.namespace:
+                msg = f'function {qname.qname!r} is already registered'
+                raise ElementPathValueError(msg)
+            elif not issubclass(token_class, XPathFunction) \
+                    or token_class.label == 'kind test':
+                msg = f'{symbol!r} name collides with {token_class!r}'
+                raise ElementPathValueError(msg)
+            else:
+                # Register a new proxy token, moving the already present token class to
+                # qualified name.
+                self.symbol_table[f'{{{token_class.namespace}}}{symbol}'] = token_class
+                class_name = f'{upper_camel_case(qname.local_name)}FunctionProxy'
+                kwargs = {
+                    'bases': (ProxyToken,),
+                    'label': 'function',
+                    'lbp': bp,
+                    'rbp': bp,
+                }
+                self._register(symbol, class_name, **kwargs)
 
+        class_name = f'{upper_camel_case(qname.qname)}ExternalFunction'
         kwargs = {
-            'class_name': class_name,
-            'symbol': symbol,
+            'lookup_name': lookup_name,
+            'bases': (ExternalFunction,),
             'namespace': namespace,
             'label': 'external function',
             'nargs': nargs,
             'lbp': bp,
             'rbp': bp,
-            'evaluate': evaluate_external_function,
-            '__module__': self.__module__,
-            '__qualname__': class_name,
-            '__return__': None
+            'sequence_types': sequence_types,
+            'callback': staticmethod(callback),
         }
-        if sequence_types:
-            # Register function signature(s)
-            kwargs['sequence_types'] = sequence_types
+        token_class = self._register(symbol, class_name, **kwargs)
+
+        if function_signatures:
+            # Register function signatures
             if self.function_signatures is self.__class__.function_signatures:
                 self.function_signatures = dict(self.__class__.function_signatures)
+            self.function_signatures.update(function_signatures)
 
-            if nargs is None:
-                pass  # pragma: no cover
-            elif isinstance(nargs, int):
-                assert len(sequence_types) == nargs + 1
-                self.function_signatures[(qname, nargs)] = 'function({}) as {}'.format(
-                    ', '.join(sequence_types[:-1]), sequence_types[-1]
-                )
-            elif nargs[1] is None:
-                assert len(sequence_types) == nargs[0] + 1
-                self.function_signatures[(qname, nargs[0])] = 'function({}, ...) as {}'.format(
-                    ', '.join(sequence_types[:-1]), sequence_types[-1]
-                )
-            else:
-                assert len(sequence_types) == nargs[1] + 1
-                for arity in range(nargs[0], nargs[1] + 1):
-                    self.function_signatures[(qname, arity)] = 'function({}) as {}'.format(
-                        ', '.join(sequence_types[:arity]), sequence_types[-1]
-                    )
-
-        token_class = cast(
-            type[XPathFunction], ABCMeta(class_name, (XPathFunction,), kwargs)
-        )
-        self.symbol_table[lookup_name] = token_class
-        self.tokenizer = None
-        return token_class
+        return cast(type[XPathFunction], token_class)
 
     def is_schema_bound(self) -> bool:
         return self.schema is not None and 'symbol_table' in self.__dict__
@@ -534,7 +445,6 @@ class XPath2Parser(XPath1Parser):
 # Remove symbols that have to be redefined for XPath 2.0.
 XPath2Parser.unregister(',')
 XPath2Parser.unregister('(')
-XPath2Parser.unregister('$')
 XPath2Parser.unregister('contains')
 XPath2Parser.unregister('lang')
 XPath2Parser.unregister('id')
