@@ -1,5 +1,5 @@
 #
-# Copyright (c), 2018-2022, SISSA (International School for Advanced Studies).
+# Copyright (c), 2018-2026, SISSA (International School for Advanced Studies).
 # All rights reserved.
 # This file is distributed under the terms of the MIT License.
 # See the file 'LICENSE' in the root directory of the present
@@ -8,16 +8,17 @@
 # @author Davide Brunato <brunato@sissa.it>
 #
 import importlib
-from collections import deque
+from collections import defaultdict, deque
 from collections.abc import Iterator
 from urllib.parse import urljoin
-from typing import cast, Any, Optional, TYPE_CHECKING, Union
+from typing import cast, Any, Optional
 from xml.etree import ElementTree
 
+import elementpath.aliases as ta
+
 from elementpath.exceptions import ElementPathRuntimeError, \
     ElementPathValueError, ElementPathKeyError
-from elementpath.aliases import NamespacesType, NsmapType, SequenceType
-from elementpath.datatypes import UntypedAtomic, AtomicType, AnyURI, QName
+from elementpath.datatypes import UntypedAtomic, AnyURI, QName
 from elementpath.namespaces import XML_NAMESPACE, XML_BASE, XSI_NIL, \
     XSD_ANY_TYPE, XSD_ANY_SIMPLE_TYPE, XSD_ANY_ATOMIC_TYPE, XSI_TYPE, \
     XML_ID, XSD_IDREF, XSD_IDREFS, XSD_UNTYPED, XSD_UNTYPED_ATOMIC, \
@@ -29,45 +30,33 @@ from elementpath.helpers import match_wildcard, is_absolute_uri
 from elementpath.decoder import get_atomic_sequence
 from elementpath.etree import etree_iter_strings, is_etree_element_instance
 
-if TYPE_CHECKING:
-    from elementpath.schema_proxy import AbstractSchemaProxy
-
-__all__ = ['TypedNodeType', 'ParentNodeType', 'ChildNodeType', 'ElementMapType',
-           'XPathNodeTree', 'XPathNode', 'NamespaceNode', 'AttributeNode', 'TextAttributeNode',
+__all__ = ['XPathNodeTree', 'XPathNode', 'NamespaceNode', 'AttributeNode', 'TextAttributeNode',
            'SchemaAttributeNode', 'TextNode', 'CommentNode', 'ProcessingInstructionNode',
            'ElementNode', 'EtreeElementNode', 'LazyElementNode', 'SchemaElementNode',
-           'DocumentNode', 'EtreeDocumentNode', 'RootNodeType', 'RootArgType']
+           'DocumentNode', 'EtreeDocumentNode']
 
 _EMPTY_NAME_PATH = f'*[Q{{{XPATH_FUNCTIONS_NAMESPACE}}}local-name()=""]'
 
 _XSD_SPECIAL_TYPES = frozenset((XSD_ANY_TYPE, XSD_ANY_SIMPLE_TYPE, XSD_ANY_ATOMIC_TYPE))
 
-TypedNodeType = Union['AttributeNode', 'ElementNode']
-TaggedNodeType = Union['ElementNode', 'CommentNode', 'ProcessingInstructionNode']
-ParentNodeType = Union['DocumentNode', 'ElementNode']
-ChildNodeType = Union['TextNode', TaggedNodeType]
-ElementMapType = dict[object, TaggedNodeType]
-FindAttrType = Optional[XsdAttributeProtocol]
-FindElemType = Optional[XsdElementProtocol]
-
 
 class XPathNodeTree:
     """
     Status of the node tree structure, shared between nodes.
     """
-    root: ParentNodeType
-    elements: ElementMapType
-    namespaces: NamespacesType
-    schema: Optional['AbstractSchemaProxy']
-    uri: Optional[str]
+    root: ta.ParentNodeType
+    elements: ta.ElementMapType
+    namespaces: ta.NamespacesType
+    schema: Optional[ta.SchemaProxyType]
+    uri: str | None
 
     __slots__ = ('root_node', 'uri', 'namespaces', 'ns_offset', 'elements', 'schema', 'position')
 
-    def __init__(self, root: ParentNodeType,
-                 uri: Optional[str] = None,
-                 namespaces: Optional[NamespacesType] = None,
-                 elements: Optional[ElementMapType] = None,
-                 schema: Optional['AbstractSchemaProxy'] = None) -> None:
+    def __init__(self, root: ta.ParentNodeType,
+                 uri: str | None = None,
+                 namespaces: Optional[ta.NamespacesType] = None,
+                 elements: Optional[ta.ElementMapType] = None,
+                 schema: Optional[ta.SchemaProxyType] = None) -> None:
 
         self.root_node = root
         self.namespaces = {} if namespaces is None else namespaces
@@ -97,33 +86,33 @@ class XPathNode:
     # XDM accessors
 
     @property
-    def attributes(self) -> Optional[list['AttributeNode']]:
+    def attributes(self) -> list['AttributeNode'] | None:
         return None
 
     @property
-    def base_uri(self) -> Optional[str]:
+    def base_uri(self) -> str | None:
         return self.parent.base_uri if self.parent is not None else None
 
-    children: Optional[list[ChildNodeType]]
+    children: list[ta.ChildNodeType] | None
 
     @property
-    def document_uri(self) -> Optional[str]:
+    def document_uri(self) -> str | None:
         return None
 
     @property
-    def is_id(self) -> Optional[bool]:
+    def is_id(self) -> bool | None:
         return None
 
     @property
-    def is_idrefs(self) -> Optional[bool]:
+    def is_idrefs(self) -> bool | None:
         return None
 
     @property
-    def namespace_nodes(self) -> Optional[list['NamespaceNode']]:
+    def namespace_nodes(self) -> list['NamespaceNode'] | None:
         return None
 
     @property
-    def nilled(self) -> Optional[bool]:
+    def nilled(self) -> bool | None:
         return None
 
     @property
@@ -131,8 +120,8 @@ class XPathNode:
         raise NotImplementedError()
 
     @property
-    def node_name(self) -> Optional[QName]:
-        name: Optional[str] = getattr(self, 'name', None)
+    def node_name(self) -> QName | None:
+        name: str | None = getattr(self, 'name', None)
         if name is None:
             return None
         elif not name.startswith('{'):
@@ -160,10 +149,10 @@ class XPathNode:
                 return QName(namespace, f"{prefix}:{local}")
         raise ElementPathKeyError(f'missing namespace prefix mapping in {self!r}')
 
-    parent: Optional[ParentNodeType]
+    parent: ta.ParentNodeType | None
 
     @property
-    def type_name(self) -> Optional[str]:
+    def type_name(self) -> str | None:
         return None
 
     @property
@@ -171,21 +160,21 @@ class XPathNode:
         raise NotImplementedError()
 
     @property
-    def typed_value(self) -> SequenceType[AtomicType]:
+    def typed_value(self) -> list[ta.AtomicType] | ta.AtomicType:
         raise NotImplementedError()
 
     @staticmethod
-    def unparsed_entity_public_id(name: str) -> Optional[str]:
+    def unparsed_entity_public_id(name: str) -> str | None:
         return None
 
     @staticmethod
-    def unparsed_entity_system_id(name: str) -> Optional[AnyURI]:
+    def unparsed_entity_system_id(name: str) -> AnyURI | None:
         return None
 
     ###
     # Other properties and methods
 
-    name: Optional[str]  # node name
+    name: str | None  # node name
     value: object        # the object wrapped in the node
     position: int        # position of the node, for document total order
 
@@ -230,26 +219,26 @@ class XPathNode:
         raise ElementPathKeyError(f'missing namespace prefix mapping in {path}')
 
     @property
-    def iter_typed_values(self) -> Iterator[AtomicType]:
+    def iter_typed_values(self) -> Iterator[ta.AtomicType]:
         raise NotImplementedError()
 
     @property
-    def is_schema_node(self) -> Optional[bool]:
+    def is_schema_node(self) -> bool | None:
         return None
 
     @property
-    def is_typed(self) -> Optional[bool]:
+    def is_typed(self) -> bool | None:
         return None
 
     @property
-    def is_extended(self) -> Optional[bool]:
+    def is_extended(self) -> bool | None:
         return None
 
     @property
-    def is_list(self) -> Optional[bool]:
+    def is_list(self) -> bool | None:
         return None
 
-    def apply_schema(self, schema: 'AbstractSchemaProxy') -> None:
+    def apply_schema(self, schema: ta.SchemaProxyType) -> None:
         """Set XSD types for elements and attribute nodes from schema proxy instance."""
         if self.parent is not None:
             self.parent.apply_schema(schema)
@@ -259,7 +248,7 @@ class XPathNode:
         if self.parent is not None:
             self.parent.clear_types()
 
-    def match_name(self, name: str, default_namespace: Optional[str] = None) -> bool:
+    def match_name(self, name: str, default_namespace: str | None = None) -> bool:
         """
         Returns `True` if the argument is matching the name of the node, `False` otherwise.
         Raises a ValueError if the argument is used, but it's in a wrong format.
@@ -271,7 +260,7 @@ class XPathNode:
         """
         return False
 
-    def get_child_position(self, child: ChildNodeType) -> int:
+    def get_child_position(self, child: ta.ChildNodeType) -> int:
         pos = 0
         if self.children:
             for c in self.children:
@@ -298,13 +287,13 @@ class NamespaceNode(XPathNode):
     :param position: the position of the node in the document.
     """
     value: str
-    parent: Optional['ElementNode']
+    parent: 'ElementNode | None'
 
     __slots__ = ()
 
     def __init__(self,
-                 prefix: Optional[str], uri: str,
-                 parent: Optional['ElementNode'] = None,
+                 prefix: str | None, uri: str,
+                 parent: 'ElementNode | None' = None,
                  position: int = 1) -> None:
         self.name = prefix
         self.value = uri
@@ -315,7 +304,7 @@ class NamespaceNode(XPathNode):
         return '%s(prefix=%r, uri=%r)' % (self.__class__.__name__, self.name, self.value)
 
     @property
-    def prefix(self) -> Optional[str]:
+    def prefix(self) -> str | None:
         return self.name
 
     @property
@@ -339,7 +328,7 @@ class NamespaceNode(XPathNode):
         return 'namespace'
 
     @property
-    def node_name(self) -> Optional[QName]:
+    def node_name(self) -> QName | None:
         return None if not self.name else QName(None, self.name)
 
     @property
@@ -358,9 +347,9 @@ class AttributeNode(XPathNode):
     """
     Base class for XPath attribute nodes, used only for type checking.
     """
-    name: Optional[str]
+    name: str | None
     parent: Optional['ElementNode']
-    xsd_type: Optional[XsdTypeProtocol]
+    xsd_type: XsdTypeProtocol | None
 
     __slots__ = ('xsd_type',)
 
@@ -373,7 +362,7 @@ class AttributeNode(XPathNode):
         return f'{self.__class__.__name__}(name={self.name!r}, value={self.value!r})'
 
     @property
-    def uri_qualified_name(self) -> Optional[str]:
+    def uri_qualified_name(self) -> str | None:
         """The URI qualified name of the attribute."""
         if not self.name:
             return self.name
@@ -402,11 +391,11 @@ class AttributeNode(XPathNode):
     def is_list(self) -> bool:
         return self.xsd_type is not None and self.xsd_type.is_list()
 
-    def match_name(self, name: str, default_namespace: Optional[str] = None) -> bool:
+    def match_name(self, name: str, default_namespace: str | None = None) -> bool:
         return self.name == name or '*' in name and match_wildcard(self.name, name)
 
     @property
-    def base_uri(self) -> Optional[str]:
+    def base_uri(self) -> str | None:
         return self.parent.base_uri if self.parent is not None else None
 
     @property
@@ -429,19 +418,16 @@ class AttributeNode(XPathNode):
         raise NotImplementedError()
 
     @property
-    def type_name(self) -> Optional[str]:
+    def type_name(self) -> str | None:
         return XSD_UNTYPED_ATOMIC if self.xsd_type is None else self.xsd_type.name
 
     @property
-    def typed_value(self) -> SequenceType[AtomicType]:
-        values = [v for v in self.iter_typed_values]
-        if len(values) == 1:
-            return values[0]
-        else:
-            return values
+    def typed_value(self) -> list[ta.AtomicType] | ta.AtomicType:
+        value = [x for x in self.iter_typed_values]
+        return value[0] if len(value) == 1 else value
 
     @property
-    def iter_typed_values(self) -> Iterator[AtomicType]:
+    def iter_typed_values(self) -> Iterator[ta.AtomicType]:
         raise NotImplementedError()
 
 
@@ -456,14 +442,14 @@ class TextAttributeNode(AttributeNode):
     """
     name: str
     value: str
-    parent: Optional['EtreeElementNode']
+    parent: 'EtreeElementNode | None'
 
     __slots__ = ()
 
     def __init__(self,
                  name: str,
                  value: str,
-                 parent: Optional['EtreeElementNode'] = None,
+                 parent: 'EtreeElementNode | None' = None,
                  position: int = 1) -> None:
 
         self.name = name
@@ -480,13 +466,13 @@ class TextAttributeNode(AttributeNode):
         return self.value
 
     @property
-    def iter_typed_values(self) -> Iterator[AtomicType]:
+    def iter_typed_values(self) -> Iterator[ta.AtomicType]:
         if self.parent is not None:
             yield from get_atomic_sequence(self.xsd_type, self.value, self.parent.nsmap)
         else:
             yield from get_atomic_sequence(self.xsd_type, self.value)
 
-    def apply_schema(self, schema: 'AbstractSchemaProxy') -> None:
+    def apply_schema(self, schema: ta.SchemaProxyType) -> None:
         if self.parent is not None:
             self.parent.apply_schema(schema)
         elif (xsd_attribute := schema.get_attribute(self.name)) is not None:
@@ -497,16 +483,16 @@ class TextAttributeNode(AttributeNode):
 
 class SchemaAttributeNode(AttributeNode):
     """A class for processing XML Schema attribute nodes."""
-    name: Optional[str]
+    name: str | None
     value: XsdAttributeProtocol
-    parent: Optional['ElementNode']
+    parent: 'ElementNode | None'
 
     __slots__ = ()
 
     def __init__(self,
                  attr: XsdAttributeProtocol,
-                 parent: Optional['ElementNode'] = None,
-                 position: int = 1):
+                 parent: 'ElementNode | None' = None,
+                 position: int = 1) -> None:
 
         self.name = attr.name
         self.value = attr
@@ -522,10 +508,10 @@ class SchemaAttributeNode(AttributeNode):
         return str(get_atomic_sequence(self.xsd_type))
 
     @property
-    def iter_typed_values(self) -> Iterator[AtomicType]:
+    def iter_typed_values(self) -> Iterator[ta.AtomicType]:
         yield from get_atomic_sequence(self.xsd_type)
 
-    def match_name(self, name: str, default_namespace: Optional[str] = None) -> bool:
+    def match_name(self, name: str, default_namespace: str | None = None) -> bool:
         if not self.name:
             return self.value.is_matching(name, default_namespace)
         elif '*' in name:
@@ -552,14 +538,14 @@ class TextNode(XPathNode):
     """
     name: None
     value: str
-    parent: Optional['ElementNode']
+    parent: 'ElementNode | None'
     children: None = None
 
     __slots__ = ()
 
     def __init__(self,
                  content: str,
-                 parent: Optional['ElementNode'] = None,
+                 parent: 'ElementNode | None' = None,
                  position: int = 1) -> None:
         self.name = None
         self.value = content
@@ -590,7 +576,7 @@ class TextNode(XPathNode):
     # Text node accessors
 
     @property
-    def base_uri(self) -> Optional[str]:
+    def base_uri(self) -> str | None:
         return self.parent.base_uri if self.parent is not None else None
 
     @property
@@ -602,11 +588,11 @@ class TextNode(XPathNode):
         return self.value
 
     @property
-    def type_name(self) -> Optional[str]:
+    def type_name(self) -> str | None:
         return XSD_UNTYPED_ATOMIC
 
     @property
-    def typed_value(self) -> SequenceType[AtomicType]:
+    def typed_value(self) -> UntypedAtomic:
         return UntypedAtomic(self.value)
 
     @property
@@ -631,8 +617,8 @@ class CommentNode(XPathNode):
     __slots__ = ()
 
     def __init__(self,
-                 content: Union[CommentType, str],
-                 parent: Union[ParentNodeType, None] = None,
+                 content: CommentType | str,
+                 parent: ta.ParentNodeType | None = None,
                  position: int = 1) -> None:
 
         self.name = None
@@ -668,7 +654,7 @@ class CommentNode(XPathNode):
         return f"/comment()[{pos}]"
 
     @property
-    def base_uri(self) -> Optional[str]:
+    def base_uri(self) -> str | None:
         return self.parent.base_uri if self.parent is not None else None
 
     @property
@@ -706,9 +692,9 @@ class ProcessingInstructionNode(XPathNode):
     __slots__ = ()
 
     def __init__(self,
-                 target: Union[str, ProcessingInstructionType],
-                 content: Optional[str] = None,
-                 parent: Optional[ParentNodeType] = None,
+                 target: str | ProcessingInstructionType,
+                 content: str | None = None,
+                 parent: ta.ParentNodeType | None = None,
                  position: int = 1) -> None:
 
         if isinstance(target, str):
@@ -757,7 +743,7 @@ class ProcessingInstructionNode(XPathNode):
         return f"/processing-instruction({self.name})[{pos}]"
 
     @property
-    def base_uri(self) -> Optional[str]:
+    def base_uri(self) -> str | None:
         return self.parent.base_uri if self.parent is not None else None
 
     @property
@@ -773,7 +759,7 @@ class ProcessingInstructionNode(XPathNode):
         return self.content
 
     @property
-    def typed_value(self) -> SequenceType[AtomicType]:
+    def typed_value(self) -> str:
         return self.content
 
     @property
@@ -791,14 +777,14 @@ class ElementNode(XPathNode):
     Base class for XPath element nodes, used only for type checking. Element nodes
     use lazy properties to diminish the average load for a tree processing.
     """
-    name: Optional[str]
+    name: str | None
     value: object
     tree: XPathNodeTree
-    children: list[ChildNodeType]
-    parent: Optional[ParentNodeType]
-    xsd_type: Optional[XsdTypeProtocol]
-    xsd_element: Optional[XsdElementProtocol]
-    _nsmap: Union[NsmapType, NamespacesType, None]
+    children: list[ta.ChildNodeType]
+    parent: ta.ParentNodeType | None
+    xsd_type: XsdTypeProtocol | None
+    xsd_element: XsdElementProtocol | None
+    _nsmap: ta.NsmapType | ta.NamespacesType | None
 
     # Lazy protected attributes
     _namespace_nodes: list[NamespaceNode]
@@ -815,30 +801,30 @@ class ElementNode(XPathNode):
     def __repr__(self) -> str:
         return '%s(elem=%r)' % (self.__class__.__name__, self.value)
 
-    def __getitem__(self, i: Union[int, slice]) -> Union[ChildNodeType, list[ChildNodeType]]:
+    def __getitem__(self, i: int | slice) -> ta.ChildNodeType | list[ta.ChildNodeType]:
         return self.children[i]
 
     def __len__(self) -> int:
         return len(self.children)
 
-    def __iter__(self) -> Iterator[ChildNodeType]:
+    def __iter__(self) -> Iterator[ta.ChildNodeType]:
         yield from self.children
 
     @property
-    def nsmap(self) -> Union[NsmapType, NamespacesType]:
+    def nsmap(self) -> ta.NsmapType | ta.NamespacesType:
         if hasattr(self.value, 'nsmap'):
-            return cast(NsmapType, self.value.nsmap)
+            return cast(ta.NsmapType, self.value.nsmap)
         elif self._nsmap is not None:
             return self._nsmap
         else:
             return self.tree.namespaces
 
     @nsmap.setter
-    def nsmap(self, nsmap: Union[NsmapType, NamespacesType]) -> None:
+    def nsmap(self, nsmap: ta.NsmapType | ta.NamespacesType) -> None:
         self._nsmap = nsmap
 
     @property
-    def uri_qualified_name(self) -> Optional[str]:
+    def uri_qualified_name(self) -> str | None:
         """The URI qualified name of the element."""
         if not self.name:
             return self.name
@@ -852,8 +838,8 @@ class ElementNode(XPathNode):
         return []
 
     @property
-    def base_uri(self) -> Optional[str]:
-        base_uri = self._uri.strip() if hasattr(self, '_uri') else None
+    def base_uri(self) -> str | None:
+        base_uri: str | None = self._uri.strip() if hasattr(self, '_uri') else None
         if self.parent is None:
             return base_uri
         elif base_uri is None:
@@ -900,19 +886,16 @@ class ElementNode(XPathNode):
         raise NotImplementedError()
 
     @property
-    def type_name(self) -> Optional[str]:
+    def type_name(self) -> str | None:
         return XSD_UNTYPED if self.xsd_type is None else self.xsd_type.name
 
     @property
-    def typed_value(self) -> SequenceType[AtomicType]:
-        values = [v for v in self.iter_typed_values]
-        if len(values) == 1:
-            return values[0]
-        else:
-            return values
+    def typed_value(self) -> list[ta.AtomicType] | ta.AtomicType:
+        value = [x for x in self.iter_typed_values]
+        return value[0] if len(value) == 1 else value
 
     @property
-    def iter_typed_values(self) -> Iterator[AtomicType]:
+    def iter_typed_values(self) -> Iterator[ta.AtomicType]:
         raise NotImplementedError()
 
     @property
@@ -920,7 +903,7 @@ class ElementNode(XPathNode):
         return self.xsd_type is not None and self.xsd_type.is_list()
 
     @property
-    def uri(self) -> Optional[str]:
+    def uri(self) -> str | None:
         return self.tree.uri
 
     @uri.setter
@@ -928,19 +911,19 @@ class ElementNode(XPathNode):
         self.tree.uri = uri
 
     @property
-    def schema(self) -> Optional['AbstractSchemaProxy']:
+    def schema(self) -> Optional[ta.SchemaProxyType]:
         return self.tree.schema
 
     @schema.setter
-    def schema(self, schema: 'AbstractSchemaProxy') -> None:
+    def schema(self, schema: ta.SchemaProxyType) -> None:
         self.tree.schema = schema
 
     @property
-    def elements(self) -> Optional[ElementMapType]:
+    def elements(self) -> Optional[ta.ElementMapType]:
         return self.tree.elements
 
     @elements.setter
-    def elements(self, elements: ElementMapType) -> None:
+    def elements(self, elements: ta.ElementMapType) -> None:
         self.tree.elements = elements
 
     @property
@@ -958,7 +941,7 @@ class ElementNode(XPathNode):
         return f"/{self.name_path}[{pos}]"
 
     @property
-    def default_namespace(self) -> Optional[str]:
+    def default_namespace(self) -> str | None:
         if None in self.nsmap:
             return self.nsmap[None]  # type: ignore
         else:
@@ -968,13 +951,13 @@ class ElementNode(XPathNode):
     def is_typed(self) -> bool:
         return self.xsd_type is not None
 
-    def apply_schema(self, schema: 'AbstractSchemaProxy') -> None:
+    def apply_schema(self, schema: ta.SchemaProxyType) -> None:
         return
 
     def clear_types(self) -> None:
         return
 
-    def match_name(self, name: str, default_namespace: Optional[str] = None) -> bool:
+    def match_name(self, name: str, default_namespace: str | None = None) -> bool:
         if self.name is None:
             return False
         elif '*' in name:
@@ -986,8 +969,7 @@ class ElementNode(XPathNode):
         else:
             return self.name == f'{{{default_namespace}}}{name}'
 
-    def get_element_node(self, elem: Union[ElementType, SchemaElemType]) \
-            -> Optional[TaggedNodeType]:
+    def get_element_node(self, elem: ElementType | SchemaElemType) -> ta.TaggedNodeType | None:
         return self.tree.elements.get(elem)
 
     def get_document_node(self, replace: bool = False, as_parent: bool = True) -> 'DocumentNode':
@@ -1049,7 +1031,7 @@ class ElementNode(XPathNode):
                 except IndexError:
                     return
 
-    def iter_descendants(self, with_self: bool = True) -> Iterator[ChildNodeType]:
+    def iter_descendants(self, with_self: bool = True) -> Iterator[ta.ChildNodeType]:
         if with_self:
             yield self
 
@@ -1086,9 +1068,9 @@ class EtreeElementNode(ElementNode):
 
     def __init__(self,
                  elem: ElementType,
-                 parent: Optional[ParentNodeType] = None,
+                 parent: ta.ParentNodeType | None = None,
                  position: int = 1,
-                 nsmap: Union[NsmapType, NamespacesType, None] = None) -> None:
+                 nsmap: ta.NsmapType | ta.NamespacesType | None = None) -> None:
 
         assert not callable(elem.tag)
         self.name = elem.tag
@@ -1162,7 +1144,7 @@ class EtreeElementNode(ElementNode):
         return self._attributes
 
     @property
-    def base_uri(self) -> Optional[str]:
+    def base_uri(self) -> str | None:
         base_uri = self.value.get(XML_BASE)
         if isinstance(base_uri, str):
             base_uri = base_uri.strip()
@@ -1192,7 +1174,7 @@ class EtreeElementNode(ElementNode):
             getattr(self.xsd_element, 'value_constraint', None) or ''
 
     @property
-    def iter_typed_values(self) -> Iterator[AtomicType]:
+    def iter_typed_values(self) -> Iterator[ta.AtomicType]:
         if self.xsd_type is None or \
                 self.xsd_type.name in _XSD_SPECIAL_TYPES or \
                 self.xsd_type.has_mixed_content():
@@ -1213,7 +1195,7 @@ class EtreeElementNode(ElementNode):
     def compat_string_value(self) -> str:
         return ''.join(etree_iter_strings(self.value))
 
-    def apply_schema(self, schema: 'AbstractSchemaProxy') -> None:
+    def apply_schema(self, schema: ta.SchemaProxyType) -> None:
         if self.tree.schema is schema and not schema.is_assertion_based():
             return
         self.tree.schema = schema
@@ -1229,7 +1211,7 @@ class EtreeElementNode(ElementNode):
 
         if schema.base_element is not None:
             self.xsd_element = schema.base_element
-            xsd_types: list[Optional[XsdTypeProtocol]] = [schema.base_element.type]
+            xsd_types: list[XsdTypeProtocol | None] = [schema.base_element.type]
             children: Iterator[Any] = iter(self)
             if schema.is_assertion_based():
                 self.xsd_type = schema.get_type(XSD_ANY_TYPE)
@@ -1239,7 +1221,7 @@ class EtreeElementNode(ElementNode):
             if hasattr(self, '_attributes'):
                 delattr(self, '_attributes')
         else:
-            root_node: ParentNodeType = self
+            root_node: ta.ParentNodeType = self
             while isinstance(root_node.parent, EtreeElementNode):
                 root_node = root_node.parent
 
@@ -1250,6 +1232,9 @@ class EtreeElementNode(ElementNode):
                 delattr(root_node, '_attributes')
 
         iterators: deque[Any] = deque()
+        element_match_cache: defaultdict[
+            int, dict[str | None, XsdElementProtocol | None]
+        ] = defaultdict(dict)
         while True:
             for node in children:
                 if not isinstance(node, EtreeElementNode):
@@ -1268,12 +1253,15 @@ class EtreeElementNode(ElementNode):
                         xsd_element = schema.get_element(node.name)
                     elif (content := xsd_types[-1].model_group) is None:
                         xsd_element = None
+                    elif node.name in (sub_cache := element_match_cache[id(content)]):
+                        xsd_element = sub_cache[node.name]
                     else:
                         for xsd_element in content.iter_elements():
                             if xsd_element.is_matching(node.name):
                                 if xsd_element.name != node.name:
                                     # a wildcard or a substitute
                                     xsd_element = schema.get_element(node.name)
+                                sub_cache[node.name] = xsd_element
                                 break
                         else:
                             xsd_element = None
@@ -1314,7 +1302,7 @@ class EtreeElementNode(ElementNode):
     def is_typed(self) -> bool:
         return self.xsd_type is not None
 
-    def match_name(self, name: str, default_namespace: Optional[str] = None) -> bool:
+    def match_name(self, name: str, default_namespace: str | None = None) -> bool:
         if '*' in name:
             return match_wildcard(self.name, name)
         elif not name:
@@ -1395,7 +1383,7 @@ class LazyElementNode(EtreeElementNode):
     """
     __slots__ = ()
 
-    def __iter__(self) -> Iterator[ChildNodeType]:
+    def __iter__(self) -> Iterator[ta.ChildNodeType]:
         if not self.children:
             if self.value.text is not None:
                 TextNode(self.value.text, self)
@@ -1413,7 +1401,7 @@ class LazyElementNode(EtreeElementNode):
 
         yield from self.children
 
-    def iter_descendants(self, with_self: bool = True) -> Iterator[ChildNodeType]:
+    def iter_descendants(self, with_self: bool = True) -> Iterator[ta.ChildNodeType]:
         if with_self:
             yield self
 
@@ -1431,15 +1419,15 @@ class SchemaElementNode(ElementNode):
     With more roots only one of them is the schema node.
     """
     value: SchemaElemType
-    ref: Optional['SchemaElementNode']
+    ref: 'SchemaElementNode | None'
 
     __slots__ = ('ref',)
 
     def __init__(self,
                  elem: SchemaElemType,
-                 parent: Optional[ParentNodeType] = None,
+                 parent: ta.ParentNodeType | None = None,
                  position: int = 1,
-                 nsmap: Union[NsmapType, NamespacesType, None] = None):
+                 nsmap: ta.NsmapType | ta.NamespacesType | None = None):
 
         self.name = elem.name
         self.value = elem
@@ -1465,7 +1453,7 @@ class SchemaElementNode(ElementNode):
 
         self.tree.elements[elem] = self
 
-    def __iter__(self) -> Iterator[ChildNodeType]:
+    def __iter__(self) -> Iterator[ta.ChildNodeType]:
         if self.ref is None:
             yield from self.children
         else:
@@ -1487,7 +1475,7 @@ class SchemaElementNode(ElementNode):
     def is_schema_node(self) -> bool:
         return True
 
-    def match_name(self, name: str, default_namespace: Optional[str] = None) -> bool:
+    def match_name(self, name: str, default_namespace: str | None = None) -> bool:
         if '*' in name:
             return match_wildcard(self.name, name)
         elif not name:
@@ -1509,8 +1497,8 @@ class SchemaElementNode(ElementNode):
         return self._attributes
 
     @property
-    def base_uri(self) -> Optional[str]:
-        base_uri = self._uri.strip() if hasattr(self, '_uri') else None
+    def base_uri(self) -> str | None:
+        base_uri: str | None = self._uri.strip() if hasattr(self, '_uri') else None
         if self.parent is None:
             return base_uri
         elif base_uri is None:
@@ -1519,9 +1507,9 @@ class SchemaElementNode(ElementNode):
             return urljoin(self.parent.base_uri or '', base_uri)
 
     @property
-    def type_name(self) -> Optional[str]:
+    def type_name(self) -> str | None:
         if (xsd_type := getattr(self.value, 'type', None)) is not None:
-            return cast(Optional[str], xsd_type.name)
+            return cast(str | None, xsd_type.name)
         return None
 
     @property
@@ -1533,7 +1521,7 @@ class SchemaElementNode(ElementNode):
         return ''
 
     @property
-    def iter_typed_values(self) -> Iterator[AtomicType]:
+    def iter_typed_values(self) -> Iterator[ta.AtomicType]:
         yield from get_atomic_sequence(self.xsd_type)
 
     def iter(self) -> Iterator[XPathNode]:
@@ -1571,7 +1559,7 @@ class SchemaElementNode(ElementNode):
                 except IndexError:
                     return
 
-    def iter_descendants(self, with_self: bool = True) -> Iterator[ChildNodeType]:
+    def iter_descendants(self, with_self: bool = True) -> Iterator[ta.ChildNodeType]:
         if with_self:
             yield self
 
@@ -1610,7 +1598,7 @@ class DocumentNode(XPathNode):
     name: None
     value: object
     parent: None
-    children: list[ChildNodeType]
+    children: list[ta.ChildNodeType]
     tree: XPathNodeTree
 
     __slots__ = ('children', 'tree')
@@ -1623,13 +1611,13 @@ class DocumentNode(XPathNode):
     def __repr__(self) -> str:
         return '%s(document=%r)' % (self.__class__.__name__, self.value)
 
-    def __getitem__(self, i: Union[int, slice]) -> Union[ChildNodeType, list[ChildNodeType]]:
+    def __getitem__(self, i: int | slice) -> ta.ChildNodeType | list[ta.ChildNodeType]:
         return self.children[i]
 
     def __len__(self) -> int:
         return len(self.children)
 
-    def __iter__(self) -> Iterator[ChildNodeType]:
+    def __iter__(self) -> Iterator[ta.ChildNodeType]:
         yield from self.children
 
     @property
@@ -1637,11 +1625,11 @@ class DocumentNode(XPathNode):
         return self.value
 
     @property
-    def uri(self) -> Optional[str]:
+    def uri(self) -> str | None:
         return self.tree.uri
 
     @property
-    def elements(self) -> dict[object, TaggedNodeType]:
+    def elements(self) -> dict[object, ta.TaggedNodeType]:
         return self.tree.elements
 
     @property
@@ -1654,7 +1642,7 @@ class DocumentNode(XPathNode):
                 return child
         raise ElementPathRuntimeError("Missing document root")
 
-    def get_element_node(self, elem: ElementProtocol) -> Optional[TaggedNodeType]:
+    def get_element_node(self, elem: ElementProtocol) -> ta.TaggedNodeType | None:
         return self.elements.get(elem)
 
     def iter(self) -> Iterator[XPathNode]:
@@ -1678,7 +1666,7 @@ class DocumentNode(XPathNode):
                 yield e
 
     def iter_descendants(self, with_self: bool = True) \
-            -> Iterator[Union['DocumentNode', ChildNodeType]]:
+            -> Iterator['DocumentNode | ta.ChildNodeType']:
         if with_self:
             yield self
 
@@ -1696,7 +1684,7 @@ class DocumentNode(XPathNode):
         else:
             return False
 
-    def apply_schema(self, schema: 'AbstractSchemaProxy') -> None:
+    def apply_schema(self, schema: ta.SchemaProxyType) -> None:
         for child in self.children:
             if isinstance(child, ElementNode):
                 child.apply_schema(schema)
@@ -1717,11 +1705,11 @@ class DocumentNode(XPathNode):
         return len(self.children) > 1 or not isinstance(self.children[0], ElementNode)
 
     @property
-    def base_uri(self) -> Optional[str]:
+    def base_uri(self) -> str | None:
         return self.uri.strip() if self.uri is not None else None
 
     @property
-    def document_uri(self) -> Optional[str]:
+    def document_uri(self) -> str | None:
         if self.uri is not None and is_absolute_uri(self.uri):
             return self.uri.strip()
         else:
@@ -1736,7 +1724,7 @@ class DocumentNode(XPathNode):
         raise NotImplementedError()
 
     @property
-    def typed_value(self) -> AtomicType:
+    def typed_value(self) -> ta.AtomicType:
         return UntypedAtomic(self.string_value)
 
     @property
@@ -1758,7 +1746,7 @@ class EtreeDocumentNode(DocumentNode):
     __slots__ = ()
 
     def __init__(self, document: DocumentType,
-                 uri: Optional[str] = None,
+                 uri: str | None = None,
                  position: int = 1) -> None:
 
         self.value = document
@@ -1807,11 +1795,3 @@ class EtreeDocumentNode(DocumentNode):
             return True
         else:
             return sum(isinstance(x, ElementNode) for x in root) != 1
-
-
-###
-# type annotation aliases
-XPathNodeType = Union[DocumentNode, NamespaceNode, AttributeNode, TextNode,
-                      ElementNode, CommentNode, ProcessingInstructionNode]
-RootNodeType = Union[DocumentNode, ElementNode]
-RootArgType = Union[DocumentType, ElementType, SchemaElemType, RootNodeType]
